//TODO: add some kind of "update" command to support pulling new oracle data
//TODO: implement oracleFileName as a cli arg
//TODO: add more robust tests
//TODO: support other list formats
const std = @import("std");

// https://github.com/Hejsil/zig-clap
const clap = @import("clap");

// https://codeberg.org/dude_the_builder/zg
const DisplayWidth = @import("DisplayWidth");

// https://github.com/AndreRenaud/PDFGen
// including the header file only works for a full build, "zig run" can't find it
const c = @cImport({
    @cInclude("pdfgen.h");
});
const print = std.debug.print;
const io = std.io;
const fs = std.fs;
// const json = std.json;
const cwd = fs.cwd();
const indexOf = std.mem.indexOf;
const expect = std.testing.expect;
const assert = std.debug.assert;
const json = std.json;

const Card = struct {
    name: []const u8 = "",          //string
    mana_cost: []const u8 = "",     //string
    cmc: f32 = 0,                   //float
    type_line: []const u8 = "",     //string
    oracle_text: []const u8 = "",   //string
    power: ?[]const u8 = null,      //coerced to string
    toughness: ?[]const u8 = null,  //coerced to string
    card_faces: ?[]Card = null,     //array of cards
    loyalty: ?[]const u8 = null,    //coerced to string
    object: []const u8 = "",        //string
    allText: [][]const u8 = undefined//
};

fn compareCards(ctx: void, lhs: Card, rhs: Card) bool {
        _ = ctx;
        return lhs.allText.len < rhs.allText.len;
}

//dimension constants and defaults
const cardWidth = 30;
const cardHeight = 32;
const minCardHeight = 5;
const pageHeight = 66;      
const pdfHeight = c.PDF_LETTER_HEIGHT;
const pdfWidth = c.PDF_LETTER_WIDTH;

const formatString = "{s: <" ++ std.fmt.digits2(cardWidth) ++ "}";
const lineFormatter = "|" ++ formatString;
const spacer: []const u8 = "|" ++ (" " ** cardWidth);
const fullWidthSpacer = (spacer ** 3) ++ "\n";
const pageWidth = fullWidthSpacer.len;

//values that may vary depending on user-supplied arguments
var oracleFileName = "oracle-data/oracle-cards-20240701090158.json";
var listFileName: []const u8 = undefined;   //the name of the file with the list of cards to print
var constantHeight = true;                   //whether or not cards must have a constant height

test "Check constants" {
    var gpa = std.heap.GeneralPurposeAllocator(.{}){};
    const paddedLine = try std.fmt.allocPrint(gpa.allocator(), lineFormatter, .{"hello> woooooooooooooooooooorl"});
    try expect(paddedLine.len == spacer.len);
}

pub fn main() !void {
    var arena = std.heap.ArenaAllocator.init(std.heap.page_allocator);
    defer arena.deinit();
    const allocator = arena.allocator();

    const params = comptime clap.parseParamsComptime(
        \\-h, --help            Display this help and exit.
        \\-l, --listFile <str>  The list of cards to print       
        \\-c, --constant <str>  "true" or "false" (case-sensitive, whether or not the height of cards may vary)
    );

    const res = clap.parse(clap.Help, &params, clap.parsers.default, .{
        .allocator = allocator,
    }) catch |err| {
        return err;
    };

    if(res.args.listFile) |name| {
        listFileName = name;
    } else {
      return error.ExpectedArgument;
    }
    if(res.args.constant) |choice| {
        constantHeight = std.mem.eql(u8, choice, "true");
    }
   
    const oracleFileSize = (try cwd.statFile(oracleFileName)).size;
    const oracleFile = try cwd.readFileAlloc(allocator, oracleFileName, oracleFileSize);
    const parsedJson = try json.parseFromSlice([]Card, allocator, oracleFile, .{ .ignore_unknown_fields = true });

    var cards = std.ArrayList(Card).init(allocator);
    const listFileSize = (try cwd.statFile(listFileName)).size;
    const listText = try cwd.readFileAlloc(allocator, listFileName, listFileSize);
    var listLines = std.mem.splitAny(u8, listText, "\n");
    while (listLines.next()) |line| {
        if (line.len < 5) break;
        const firstSpace = indexOf(u8, line, " ").?;
        const cardName = line[firstSpace + 1 .. indexOf(u8, line, "(").? - 1];
        const countString = line[0..firstSpace];
        const printCount = try std.fmt.parseInt(u8, countString, 10);
        for(0..printCount) |_| {
            try cards.append(Card{.name = cardName});
        }
    }

    for (parsedJson.value) |*cardObj| {
        for(cards.items) |*cardInfo| {
            if (std.mem.eql(u8, cardInfo.name, cardObj.name)) {
                cardInfo.allText = try card(cardObj.*, allocator);
            }
        }
    }

    std.mem.sort(Card, cards.items, {}, compareCards);

    var allPrinted = std.ArrayList([]const u8).init(allocator);
    const pdf_doc: *c.pdf_doc = @ptrCast(c.pdf_create(pdfWidth, pdfHeight, &c.pdf_info{ .creator = ("My Software" ++ " " ** 53).* }).?);

    var rowToPrint = std.ArrayList(Card).init(allocator);
    for (cards.items) |cardText| {
        try rowToPrint.append(cardText);
        if (rowToPrint.items.len >= 3) {
            try printRow(allocator, rowToPrint.items[0..3].*, &allPrinted);
            rowToPrint.clearAndFree();
        }
    } else {
        try printRow(allocator, [3]Card{rowToPrint.items[0], Card{}, Card{}}, &allPrinted);
        rowToPrint.clearAndFree();
    }
    var page = c.pdf_append_page(pdf_doc);
    _ = c.pdf_set_font(pdf_doc, "Courier");
    for (allPrinted.items, 1..) |text, rowNum| {
        const pageRelative = rowNum % pageHeight;
        const pageOffset: f32 = 12 * @as(f32, @floatFromInt(pageRelative));
        if (pageRelative == 0) {
            page = c.pdf_append_page(pdf_doc);
        }
        _ = c.pdf_add_text(pdf_doc, page, try std.mem.Allocator.dupeZ(allocator, u8, text), 8, 10, pdfHeight - pageOffset, c.PDF_BLACK);
    }

    _ = c.pdf_save(pdf_doc, "output.pdf");
    _ = c.pdf_destroy(pdf_doc);
}

fn card(
    cardObj: Card,
    allocator: std.mem.Allocator,
) ![][]const u8 {
    const isFace= std.mem.eql(u8, cardObj.object, "card_face");
    var cardText = std.ArrayList([]const u8).init(allocator);

    var fullUnformattedText = std.ArrayList(u8).init(allocator);

    var oracleIterator = std.mem.splitAny(u8, cardObj.oracle_text, "()");
    var oracleBuf = std.ArrayList(u8).init(allocator);
    var oracleIdx: usize = 0;
    while(oracleIterator.next()) |slice| : (oracleIdx += 1) {
        if(oracleIdx % 2 != 0) continue;
        try oracleBuf.appendSlice(slice);
    }
    var finalOracleText: [30 * 100]u8 = undefined;
    _ = std.mem.replace(u8, oracleBuf.items, cardObj.name, "~", &finalOracleText);

    if (cardObj.card_faces == null or isFace) {
        try fullUnformattedText.appendSlice(try std.mem.concat(allocator, u8, &[_][]const u8{
            cardObj.name,
            if (cardObj.mana_cost.len > 0) " " else "",
            cardObj.mana_cost,
            try std.fmt.allocPrint(allocator, " ({s}) >> ", .{cardObj.type_line}),
            &finalOracleText,
            if(cardObj.power) |_| try std.fmt.allocPrint(allocator, "({s}/{s})", .{cardObj.power.?, cardObj.toughness.?}) else "",
            if(cardObj.loyalty) |_| try std.fmt.allocPrint(allocator, "[{s}]", .{cardObj.loyalty.?}) else ""
        }));
    }

    if (cardObj.card_faces) |faces| {
        for (faces, 0..) |*face, idx| {
            const faceText = try card(face.*, allocator);
            try fullUnformattedText.appendSlice(std.mem.trim(u8, try std.mem.join(allocator, " ", faceText), "\n"));
            if (idx == 0) try fullUnformattedText.appendSlice("// ");
        }
    }

    var line = std.ArrayList(u8).init(allocator);
    var wordIterator = std.mem.splitAny(u8, fullUnformattedText.items, &.{'\n',' ',170});
    while (wordIterator.next()) |word| {
        if (line.items.len + word.len + 1 >= cardWidth) {
            try cardText.append(try line.toOwnedSlice());
            line.clearAndFree();
        }
        if(word.len > 0) {
            try line.appendSlice(word);
            try line.append(' ');
            assert(line.items.len < 30);
        }
    } else {
        try cardText.append(try line.toOwnedSlice());
    }
    if(!isFace and constantHeight) {
        assert(cardText.items.len <= cardHeight);
        try cardText.appendNTimes(" " ** (cardWidth - 2), cardHeight - cardText.items.len);
    } else if(!isFace) {
        try cardText.append(" " ** (cardWidth - 2));
    }
    assert(cardText.items.len <= cardHeight);
    return try cardText.toOwnedSlice();
}

fn optionalAccess(slice: anytype, index: usize) []const u8 {
   if(slice.len > index) return slice[index];
   return "";
}

fn printRow(allocator: std.mem.Allocator, cards: [3]Card, allPrinted: *std.ArrayList([]const u8)) !void {
    var lines = std.ArrayList([]const u8).init(allocator);

    const endIndex = std.mem.max(u64, &.{
        cards[0].allText.len, 
        cards[1].allText.len, 
        cards[2].allText.len
    });

    for(0..endIndex) |idx| {
        try lines.append(
            try std.fmt.allocPrint(allocator, lineFormatter ** 3 ++ "\n", .{
                optionalAccess(cards[0].allText, idx), 
                optionalAccess(cards[1].allText, idx),
                optionalAccess(cards[2].allText, idx)
            })
         );
    }

    const rowHeight = lines.items.len;
    while (pageHeight - (allPrinted.items.len % pageHeight) <= rowHeight) {
        assert(rowHeight <= pageHeight);
        try allPrinted.append(fullWidthSpacer);
    }
    try allPrinted.appendSlice(lines.items);
}
